@@ -1,14 +1,18 @@
-# 
+#!/usr/bin/env python3
+
+#
 # (c) Simon Marlow 2002
 #
 
 from __future__ import print_function
 
+import signal
 import sys
 import os
 import string
 import getopt
-import platform
+import shutil
+import tempfile
 import time
 import re
 
@@ -20,11 +24,6 @@ import re
 # So we import it here first, so that the testsuite doesn't appear to fail.
 import subprocess
 
-PYTHON3 = sys.version_info >= (3, 0)
-if PYTHON3:
-    print("*** WARNING: running testsuite using Python 3.\n"
-          "*** Python 3 support is experimental. See Trac #9184.")
-
 from testutil import *
 from testglobals import *
 
@@ -36,6 +35,9 @@ os.environ['TERM'] = 'vt100'
 global config
 config = getConfig() # get it from testglobals
 
+def signal_handler(signal, frame):
+        stopNow()
+
 # -----------------------------------------------------------------------------
 # cmd-line options
 
@@ -105,45 +107,16 @@ for opt,arg in opts:
         config.threads = int(arg)
         config.use_threads = 1
 
-    if opt == '--check-files-written':
-        config.check_files_written = True
-
     if opt == '--skip-perf-tests':
         config.skip_perf_tests = True
 
     if opt == '--verbose':
-        if arg not in ["0","1","2","3","4"]:
-            sys.stderr.write("ERROR: requested verbosity %s not supported, use 0,1,2,3 or 4" % arg)
+        if arg not in ["0","1","2","3","4","5"]:
+            sys.stderr.write("ERROR: requested verbosity %s not supported, use 0,1,2,3,4 or 5" % arg)
             sys.exit(1)
         config.verbose = int(arg)
 
 
-if config.use_threads == 1:
-    # Trac #1558 says threads don't work in python 2.4.4, but do
-    # in 2.5.2. Probably >= 2.5 is sufficient, but let's be
-    # conservative here.
-    # Some versions of python have things like '1c1' for some of
-    # these components (see trac #3091), but int() chokes on the
-    # 'c1', so we drop it.
-    (maj, min, pat) = platform.python_version_tuple()
-    # We wrap maj, min, and pat in str() to work around a bug in python
-    # 2.6.1
-    maj = int(re.sub('[^0-9].*', '', str(maj)))
-    min = int(re.sub('[^0-9].*', '', str(min)))
-    pat = int(re.sub('[^0-9].*', '', str(pat)))
-    if (maj, min) < (2, 6):
-        print("Python < 2.6 is not supported")
-        sys.exit(1)
-    # We also need to disable threads for python 2.7.2, because of
-    # this bug: http://bugs.python.org/issue13817
-    elif (maj, min, pat) == (2, 7, 2):
-        print("Warning: Ignoring request to use threads as python version is 2.7.2")
-        print("See http://bugs.python.org/issue13817 for details.")
-        config.use_threads = 0
-    if windows: # See Trac ticket #10510.
-        print("Warning: Ignoring request to use threads as running on Windows")
-        config.use_threads = 0
-
 config.cygwin = False
 config.msys = False
 
@@ -163,19 +136,23 @@ if windows:
 # Try to use UTF8
 if windows:
     import ctypes
-    # Windows Python provides windll, mingw python provides cdll.
-    if hasattr(ctypes, 'windll'):
-        mydll = ctypes.windll
+    # Windows and mingw* Python provide windll, msys2 python provides cdll.
+    if hasattr(ctypes, 'WinDLL'):
+        mydll = ctypes.WinDLL
     else:
-        mydll = ctypes.cdll
+        mydll = ctypes.CDLL
 
     # This actually leaves the terminal in codepage 65001 (UTF8) even
     # after python terminates. We ought really remember the old codepage
     # and set it back.
-    if mydll.kernel32.SetConsoleCP(65001) == 0:
+    kernel32 = mydll('kernel32.dll')
+    if kernel32.SetConsoleCP(65001) == 0:
         raise Exception("Failure calling SetConsoleCP(65001)")
-    if mydll.kernel32.SetConsoleOutputCP(65001) == 0:
+    if kernel32.SetConsoleOutputCP(65001) == 0:
         raise Exception("Failure calling SetConsoleOutputCP(65001)")
+
+    # register the interrupt handler
+    signal.signal(signal.SIGINT, signal_handler)
 else:
     # Try and find a utf8 locale to use
     # First see if we already have a UTF8 locale
@@ -209,7 +186,7 @@ from testlib import *
 # On Windows we need to set $PATH to include the paths to all the DLLs
 # in order for the dynamic library tests to work.
 if windows or darwin:
-    pkginfo = getStdout([config.ghc_pkg, 'dump'])
+    pkginfo = str(getStdout([config.ghc_pkg, 'dump']))
     topdir = config.libdir
     if windows:
         mingw = os.path.join(topdir, '../mingw/bin')
@@ -240,12 +217,6 @@ if windows or darwin:
 global testopts_local
 testopts_local.x = TestOptions()
 
-if config.use_threads:
-    t.lock = threading.Lock()
-    t.thread_pool = threading.Condition(t.lock)
-    t.lockFilesWritten = threading.Lock()
-    t.running_threads = 0
-
 # if timeout == -1 then we try to calculate a sensible value
 if config.timeout == -1:
     config.timeout = int(read_no_crs(config.top + '/timeout/calibrate.out'))
@@ -258,7 +229,7 @@ print('Timeout is ' + str(config.timeout))
 if config.rootdirs == []:
     config.rootdirs = ['.']
 
-t_files = findTFiles(config.rootdirs)
+t_files = list(findTFiles(config.rootdirs))
 
 print('Found', len(t_files), '.T files...')
 
@@ -270,28 +241,54 @@ t.start_time = time.localtime()
 print('Beginning test run at', time.strftime("%c %Z",t.start_time))
 
 sys.stdout.flush()
-if PYTHON3:
-    # in Python 3, we output text, which cannot be unbuffered
-    sys.stdout = os.fdopen(sys.__stdout__.fileno(), "w")
+# we output text, which cannot be unbuffered
+sys.stdout = os.fdopen(sys.__stdout__.fileno(), "w")
+
+if config.local:
+    tempdir = ''
 else:
-    # set stdout to unbuffered (is this the best way to do it?)
-    sys.stdout = os.fdopen(sys.__stdout__.fileno(), "w", 0)
+    # See note [Running tests in /tmp]
+    tempdir = tempfile.mkdtemp('', 'ghctest-')
+
+    # opts.testdir should be quoted when used, to make sure the testsuite
+    # keeps working when it contains backward slashes, for example from
+    # using os.path.join. Windows native and mingw* python
+    # (/mingw64/bin/python) set `os.path.sep = '\\'`, while msys2 python
+    # (/bin/python, /usr/bin/python or /usr/local/bin/python) sets
+    # `os.path.sep = '/'`.
+    # To catch usage of unquoted opts.testdir early, insert some spaces into
+    # tempdir.
+    tempdir = os.path.join(tempdir, 'test   spaces')
+
+def cleanup_and_exit(exitcode):
+    if config.cleanup and tempdir:
+        shutil.rmtree(tempdir, ignore_errors=True)
+    exit(exitcode)
 
 # First collect all the tests to be run
+t_files_ok = True
 for file in t_files:
     if_verbose(2, '====> Scanning %s' % file)
-    newTestDir(os.path.dirname(file))
+    newTestDir(tempdir, os.path.dirname(file))
     try:
-        exec(open(file).read())
-    except Exception:
-        print('*** framework failure: found an error while executing ', file, ':')
-        t.n_framework_failures = t.n_framework_failures + 1
-        traceback.print_exc()
+        with io.open(file, encoding='utf8') as f:
+            src = f.read()
 
-if config.only:
-    # See Note [Mutating config.only]
-    sys.stderr.write("ERROR: tests not found: {0}\n".format(list(config.only)))
-    sys.exit(1)
+        exec(src)
+    except Exception as e:
+        traceback.print_exc()
+        framework_fail(file, '', str(e))
+        t_files_ok = False
+
+for name in config.only:
+    if t_files_ok:
+        # See Note [Mutating config.only]
+        framework_fail(name, '', 'test not found')
+    else:
+        # Let user fix .T file errors before reporting on unfound tests.
+        # The reson the test can not be found is likely because of those
+        # .T file errors.
+        pass
 
 if config.list_broken:
     global brokens
@@ -300,32 +297,69 @@ if config.list_broken:
     print(' '.join(map (lambda bdn: '#' + str(bdn[0]) + '(' + bdn[1] + '/' + bdn[2] + ')', brokens)))
     print('')
 
-    if t.n_framework_failures != 0:
-        print('WARNING:', str(t.n_framework_failures), 'framework failures!')
+    if t.framework_failures:
+        print('WARNING:', len(framework_failures), 'framework failures!')
         print('')
 else:
+    # completion watcher
+    watcher = Watcher(len(parallelTests))
+
     # Now run all the tests
-    if config.use_threads:
-        t.running_threads=0
     for oneTest in parallelTests:
         if stopping():
             break
-        oneTest()
-    if config.use_threads:
-        t.thread_pool.acquire()
-        while t.running_threads>0:
-            t.thread_pool.wait()
-        t.thread_pool.release()
+        oneTest(watcher)
+
+    # wait for parallel tests to finish
+    if not stopping():
+        watcher.wait()
+
+    # Run the following tests purely sequential
     config.use_threads = False
     for oneTest in aloneTests:
         if stopping():
             break
-        oneTest()
-        
+        oneTest(watcher)
+
+    # flush everything before we continue
+    sys.stdout.flush()
+
     summary(t, sys.stdout, config.no_print_summary)
 
     if config.summary_file != '':
-        summary(t, open(config.summary_file, 'w'))
+        with open(config.summary_file, 'w') as file:
+            summary(t, file)
 
-sys.exit(0)
+cleanup_and_exit(0)
 
+# Note [Running tests in /tmp]
+#
+# Use LOCAL=0 to run tests in /tmp, to catch tests that use files from
+# the source directory without copying them to the test directory first.
+#
+# As an example, take a run_command test with a Makefile containing
+# `$(TEST_HC) ../Foo.hs`. GHC will now create the output files Foo.o and
+# Foo.hi in the source directory. There are 2 problems with this:
+# * Output files in the source directory won't get cleaned up automatically.
+# * Two tests might (over)write the same output file.
+#
+# Tests that only fail when run concurrently with other tests are the
+# worst, so we try to catch them early by enabling LOCAL=0 in validate.
+#
+# Adding -outputdir='.' to TEST_HC_OPTS would help a bit, but it requires
+# making changes to quite a few tests. The problem is that
+# `$(TEST_HC) ../Foo.hs -outputdir=.` with Foo.hs containing
+# `module Main where` does not produce Foo.o, as it would without
+# -outputdir, but Main.o. See [1].
+#
+# Using -outputdir='.' is not foolproof anyway, since it does not change
+# the destination of the final executable (Foo.exe).
+#
+# Another hardening method that could be tried is to `chmod -w` the
+# source directory.
+#
+# By default we set LOCAL=1, because it makes it easier to inspect the
+# test directory while working on a new test.
+#
+# [1]
+# https://downloads.haskell.org/~ghc/8.0.1/docs/html/users_guide/separate_compilation.html#output-files
